/**
 * File:   cPad.h
 * Author: Benjamin Pickhardt (hackCraft.de)
 *
 * Created on March 28, 2008, 10:23 PM
 *
 * Provides Uniform GamePad-Input in a Common Fashion.
 * Possibly-Mapped-Input-Sources:
 * - Keyboard(+Mouse), Real-GamePad, AI-Controll, Remote-Controll (&Proxies).
 * - Input recording and playback for demoloops.
 */

#if !defined(PAD_INC)
#define PAD_INC

// for memcpy:
#include <string>

/**
 * Emulation of a hardwired "universal" gamepad.
 * It's Structure is like that of available Gamepads.
 */
class cPad {
private:
    char axes[4];
    unsigned short buttons;
public:

    /**
     * Identifiers for single (analouge) Axes:
     * UD1, LR1: Primary stick.
     * UD2, LR2: Secondary stick.
     */
    enum Axes {
        AX_LR1, AX_UD1, AX_UD2, AX_LR2
    };

    /**
     * Identifiers for single Buttons with groupings:
     * HU, HD, HL, HR: CoolieHat or digital axes as of yesterdays.
     * DELTA, CROSS, SQUARE, CIRCLE: Primary action buttons (up,dn,lt,rt).
     * L1, L2: Left hand side buttons (top and bottom) [selections/camera/trigger].
     * R1, R2: Right hand side buttons (top and bottom) [selections/camera/trigger].
     */
    enum Buttons {
        BT_SQUARE, BT_DELTA, BT_CROSS, BT_CIRCLE, BT_L1, BT_L2, BT_R1, BT_R2, BT_SELECT, BT_START, BT_J1B, BT_J2B, BT_HU, BT_HD, BT_HL, BT_HR
    };

    /**
     * A new instance of a pad will be initialised to all zeros.
     */
    cPad() {
        reset();
    }

    /**
     * Methods for setting and resetting axis and button value.
     * Use the enum-constants of "Axes" and "Buttons" to access
     * to individual axes and buttons.
     *
     * Ranges in float for Axis are [-1,+1], whereas for int it is [-127,+127].
     */
    void reset() {
        axes[0] = axes[1] = axes[2] = axes[3] = 0;
        buttons = 0;
    };

    void setAxis(Axes axis, int value) {
        this->axes[axis] = (value < -127) ? -127 : (value > +127) ? +127 : value;
    };

    void setAxis(Axes axis, float value) {
        this->axes[axis] = (char) (127.0f * ((value < -1.0f) ? -1.0f : (value > +1.0f) ? +1.0f : value));
    };

    void setButton(Buttons button, bool enable) {
        if (enable) this->buttons |= (1UL << button);
        else this->buttons &= ~(1UL << button);
    };

    /**
     * Use the enum-constants of "Axes" and "Buttons" to access
     * to individual axes and buttons.
     *
     * Ranges in float for Axis are [-1,+1], whereas for int it is [-127,+127].
     */
    float getAxis(Axes axis) {
        return (((float) this->axes[axis]) / 127.0f);
    };

    int getButton(Buttons button) {
        if ((this->buttons & (1UL << button)) != 0) return 1;
        return 0;
    };

    /**
     * Prints the current pad configuration to stdout.
     */
    void print() {
        printf("Pad: J1(%1.2f, %1.2f) J2(%1.2f, %1.2f) ABXY(%i, %i, %i, %i) L1/R1(%i, %i) L2/R2(%i, %i) SEL/ST(%i, %i) J1B(%i) J2B(%i) HUD(%i, %i, %i, %i)\n",
                getAxis(AX_UD1), getAxis(AX_LR1),
                getAxis(AX_UD2), getAxis(AX_LR2),
                getButton(BT_SQUARE), getButton(BT_DELTA), getButton(BT_CROSS), getButton(BT_CIRCLE),
                getButton(BT_L1), getButton(BT_R1),
                getButton(BT_L2), getButton(BT_R2),
                getButton(BT_SELECT), getButton(BT_START),
                getButton(BT_J1B), getButton(BT_J2B),
                getButton(BT_HU), getButton(BT_HD), getButton(BT_HL), getButton(BT_HR)
                );
    }

    /**
     * Input recording.
     * Takes a pointer to a buffer,
     * puts in the current pad configuration,
     * increments the buffer pointer by the size of the configuration
     * and returns that incremented pointer to the buffer
     * so that it can be used for the next call.
     */
    char* serialise(char *buffer) {
        int size = 8;
        std::memcpy(buffer, this, size);
        return (buffer + size);
    }

    /**
     * Input playback for demo replay
     * Just like input recording but reads a pad
     * configuration from the buffer instead.
     */
    char* deserialise(char* buffer) {
        int size = 8;
        std::memcpy(this, buffer, size);
        return (buffer + size);
    }
};

#endif
